from sqlalchemy import Column, Integer, String, ForeignKey, Float
from sqlalchemy.orm import relationship, validates

from huoguoml.server.entity import Base


class MLModelORM(Base):
    __tablename__ = "ml_models"

    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, index=True)
    version = Column(String, index=True)
    tag = Column(Integer)
    creation_time = Column(Float)
    last_modification = Column(Float)

    run_id = Column(Integer, ForeignKey('runs.id'), unique=True)
    run = relationship("RunORM", back_populates="ml_model")

    ml_services = relationship("MLServiceORM",
                               back_populates="model")

    @validates('name')
    def convert_upper(self, key, value):
        self.lower = value.lower()
        return self.lower



import time
from typing import List, Dict, Optional

from huoguoml.schema.run import RunIn, Run
from huoguoml.server.entity.experiment import ExperimentORM
from huoguoml.server.entity.run import RunORM
from huoguoml.server.repository import Repository


class RunRepository(Repository):

    def __init__(self, database_url: str, connect_args: Dict):
        super(RunRepository, self).__init__(database_url=database_url, connect_args=connect_args)

    def create_run(self, run_in: RunIn) -> RunORM:
        session = self.Session()
        experiment = session.query(ExperimentORM).filter_by(name=run_in.experiment_name).first()

        if not experiment:
            experiment = ExperimentORM(
                description="",
                name=run_in.experiment_name)
            session.add(experiment)
            session.commit()
            session.refresh(experiment)

        run = RunORM(
            run_nr=len(experiment.runs) + 1,
            creation_time=time.time(),
            last_modification=time.time(),
            finish_time=-1,
            duration=-1,
            description="",
            parameters={},
            metrics={},
            tags={},
            model_definition=None,
            **run_in.dict()
        )

        session.add(run)
        session.commit()
        session.refresh(run)
        return run

    def get_run(self, run_id: int) -> Optional[RunORM]:
        session = self.Session()
        return session.query(RunORM).filter_by(id=run_id).first()

    def update_run(self, run_id: int, run: Run) -> Optional[RunORM]:
        session = self.Session()
        run_orm = session.query(RunORM).filter_by(id=run_id).first()

        if run_orm:
            for field, field_value in run.dict(exclude={"id", "run_nr", "experiment_name", "ml_model"}).items():
                setattr(run_orm, field, field_value)
            run_orm.last_modification = time.time()
            session.commit()
            session.refresh(run_orm)
            return run_orm

    def get_runs_by_experiment_name_and_run_nrs(self, experiment_name: str, run_nrs: List[int]) -> List[RunORM]:
        session = self.Session()
        return session.query(RunORM).filter(experiment_name == RunORM.experiment_name,
                                            RunORM.run_nr.in_(run_nrs)).all()

    def get_runs(self) -> List[RunORM]:
        session = self.Session()
        return session.query(RunORM).all()



import getpass
import time
from abc import ABC
from typing import Optional, List

import requests
from starlette.datastructures import UploadFile

from huoguoml.schema.run import RunStatus, Run, RunIn
from huoguoml.server import RunService
from huoguoml.util.string import concat_uri


class HuoguoMLRun(object):

    def __init__(self,
                 run: Run):
        self._run = run

    def log_model(self,
                  model_type: str,
                  **kwargs):
        if self._run.model_definition:
            raise FileExistsError("Model already logged")

        if model_type == "tensorflow":
            from huoguoml.tracking.tensorflow import log_model
            model_definition, model_files = log_model(**kwargs)
        else:
            raise NotImplementedError("HuoguoML does not support {} at the moment".format(model_type))

        self._run.model_definition = model_definition
        self._update_experiment_run(model_files)

    def log_parameter(self, parameter_name: str, parameter_value: str):
        self._run.parameters[parameter_name] = parameter_value
        self._update_experiment_run()

    def log_metric(self, metric_name: str, metric_value: str):
        self._run.metrics[metric_name] = metric_value
        self._update_experiment_run()

    def log_tag(self, tag_name: str, tag_value: str):
        self._run.tags[tag_name] = tag_value
        self._update_experiment_run()

    def end_experiment_run(self, failed=False):
        self._run.finish_time = time.time()
        self._run.duration = self._run.finish_time - self._run.creation_time

        if failed:
            self._run.status = RunStatus.failed.value
        else:
            self._run.status = RunStatus.completed.value

        self._update_experiment_run()

    def _update_experiment_run(self, files: Optional[List] = None):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_value:
            self.end_experiment_run(failed=True)
        else:
            self.end_experiment_run(failed=False)


class HuoguoMLAPIRun(HuoguoMLRun, ABC):

    def __init__(self, experiment_name: str, server_uri: str):
        self.server_uri = server_uri

        run_in = RunIn(experiment_name=experiment_name, author=getpass.getuser())
        run_res = requests.post(
            concat_uri(self.server_uri, "api", "runs"),
            json=run_in.dict())

        super(HuoguoMLAPIRun, self).__init__(run=Run.parse_raw(run_res.text))

    def _update_experiment_run(self, files: Optional[List] = None):
        api = concat_uri(self.server_uri, "api", "runs", str(self._run.id))
        if files:
            requests.put(concat_uri(api, "files"), files=files)

        requests.put(
            api,
            json=self._run.dict())


class HuoguoMLLocalRun(HuoguoMLRun, ABC):

    def __init__(self, experiment_name: str, artifact_dir: str):
        self.service = RunService(artifact_dir=artifact_dir)

        run_in = RunIn(experiment_name=experiment_name, author=getpass.getuser())
        run_orm = self.service.create_run(run_in=run_in)

        super(HuoguoMLLocalRun, self).__init__(run=Run.from_orm(run_orm))

    def _update_experiment_run(self, files: Optional[List] = None):
        if files:
            files = [UploadFile(filename=filename, file=file) for _, (filename, file) in files]
            self.service.update_or_create_run_files(run_id=self._run.id, files=files)

        self.service.update_run(run_id=self._run.id, run=self._run)



"""
The huoguoml.repository module contains the classes for all database operations
"""

from typing import Dict

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from huoguoml.server.entity import Base


class Repository(object):
    """The Repository object is responsible for the connection to the database.
    """

    def __init__(self, database_url: str, connect_args: Dict):
        engine = create_engine(
            database_url, connect_args=connect_args
        )
        session_factory = sessionmaker(bind=engine)
        self.Session = scoped_session(session_factory)
        Base.metadata.create_all(bind=engine)



from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship

from huoguoml.server.entity import Base


class MLServiceORM(Base):
    __tablename__ = "ml_services"

    id = Column(Integer, primary_key=True, index=True)
    host = Column(String)
    port = Column(Integer)
    model_rule = Column(String)
    model_version = Column(String)
    model_name = Column(String)

    model_id = Column(String, ForeignKey('ml_models.id'), nullable=False)
    model = relationship("MLModelORM", back_populates="ml_services")



from typing import List, Optional

from huoguoml.schema.experiment import Experiment
from huoguoml.server.entity.experiment import ExperimentORM
from huoguoml.server.entity.run import RunORM
from huoguoml.server.repository.experiment import ExperimentRepository
from huoguoml.server.service import Service


class ExperimentService(Service):

    def __init__(self, artifact_dir: str):
        super(ExperimentService, self).__init__(artifact_dir=artifact_dir)

        self.repository = ExperimentRepository(database_url=self.database_url,
                                               connect_args=self.connect_args)

    def get_experiments(self) -> List[ExperimentORM]:
        return self.repository.get_experiments()

    def get_experiment(self, experiment_name: str) -> Optional[ExperimentORM]:
        return self.repository.get_experiment(experiment_name=experiment_name)

    def get_experiment_run(self, experiment_name: str, experiment_run_nr: int) -> Optional[RunORM]:
        return self.repository.get_experiment_run(experiment_name=experiment_name,
                                                  experiment_run_nr=experiment_run_nr)

    def update_experiment(self, experiment_name: str, experiment: Experiment) -> Optional[ExperimentORM]:
        experiment_orm = self.repository.update_experiment(experiment_name=experiment_name, experiment=experiment)
        return experiment_orm



from sqlalchemy import Column, Integer, String, ForeignKey, Float, PickleType
from sqlalchemy.orm import relationship, validates

from huoguoml.server.entity import Base


class RunORM(Base):
    __tablename__ = "runs"

    id = Column(Integer, primary_key=True, index=True, unique=True)
    run_nr = Column(Integer, index=True)

    creation_time = Column(Float)
    finish_time = Column(Float)
    last_modification = Column(Float)
    duration = Column(Float)

    status = Column(Integer)
    description = Column(String)
    author = Column(String)

    parameters = Column(PickleType)
    metrics = Column(PickleType)
    tags = Column(PickleType)
    model_definition = Column(PickleType)

    experiment = relationship("ExperimentORM", back_populates="runs")
    experiment_name = Column(String, ForeignKey("experiments.name"))

    ml_model = relationship("MLModelORM", uselist=False, back_populates="run")

    @validates('experiment_name')
    def convert_upper(self, key, value):
        self.lower = value.lower()
        return self.lower



from typing import List, Dict, Optional

from huoguoml.schema.experiment import Experiment
from huoguoml.server.entity.experiment import ExperimentORM
from huoguoml.server.entity.run import RunORM
from huoguoml.server.repository import Repository


class ExperimentRepository(Repository):

    def __init__(self, database_url: str, connect_args: Dict):
        super(ExperimentRepository, self).__init__(database_url=database_url, connect_args=connect_args)

    def get_experiments(self) -> List[ExperimentORM]:
        session = self.Session()
        return session.query(ExperimentORM).all()

    def get_experiment(self, experiment_name: str) -> Optional[ExperimentORM]:
        session = self.Session()
        experiment = session.query(ExperimentORM).filter_by(name=experiment_name).first()
        return experiment

    def get_experiment_run(self, experiment_name: str, experiment_run_nr: int) -> Optional[RunORM]:
        session = self.Session()
        return session.query(RunORM).filter_by(run_nr=experiment_run_nr, experiment_name=experiment_name
                                               ).first()

    def update_experiment(self, experiment_name: str, experiment: Experiment) -> Optional[ExperimentORM]:
        session = self.Session()
        experiment_orm = session.query(ExperimentORM).filter_by(name=experiment_name, id=experiment.id).first()
        if experiment_orm:
            for field, field_value in experiment.dict(exclude_unset=True, exclude={"id", "runs"}).items():
                setattr(experiment_orm, field, field_value)

            session.commit()
            session.refresh(experiment_orm)
            return experiment_orm



from typing import Dict, Optional

from sqlalchemy import desc

from huoguoml.schema.ml_model import MLModelTag
from huoguoml.schema.ml_service import MLServiceIn
from huoguoml.server.entity.ml_model import MLModelORM
from huoguoml.server.entity.ml_service import MLServiceORM
from huoguoml.server.repository import Repository


class MLServiceRepository(Repository):

    def __init__(self, database_url: str, connect_args: Dict):
        super(MLServiceRepository, self).__init__(database_url=database_url, connect_args=connect_args)

    def create_or_update_ml_service(self, ml_service_in: MLServiceIn) -> Optional[MLServiceORM]:
        session = self.Session()

        ml_model = None
        if ml_service_in.model_rule == "production" or ml_service_in.model_rule == "staging":
            ml_model = session.query(MLModelORM).filter_by(name=ml_service_in.model_name,
                                                           tag=MLModelTag[ml_service_in.model_rule]).first()
        elif ml_service_in.model_rule == "latest":
            ml_model = session.query(MLModelORM).filter_by(name=ml_service_in.model_name).order_by(
                desc(MLModelORM.version)).first()

        if ml_model:
            ml_service = session.query(MLServiceORM).filter_by(host=ml_service_in.host,
                                                               port=ml_service_in.port).first()
            if ml_service:
                ml_service.model_name = ml_model.name
                ml_service.model_version = ml_model.version
                ml_service.model = ml_model
            else:
                ml_service = MLServiceORM(
                    host=ml_service_in.host,
                    port=ml_service_in.port,
                    model=ml_model,
                    model_name=ml_model.name,
                    model_version=ml_model.version,
                    model_rule=ml_service_in.model_rule,
                )
                session.add(ml_service)
            session.commit()
            session.refresh(ml_service)
            return ml_service

    def get_ml_services(self):
        session = self.Session()
        return session.query(MLServiceORM).all()

    def get_ml_service(self, service_id: int) -> Optional[MLServiceORM]:
        session = self.Session()
        return session.query(MLServiceORM).filter_by(id=service_id).first()



from typing import List, Any

from fastapi import APIRouter, HTTPException

from huoguoml.schema.experiment import Experiment
from huoguoml.schema.run import Run
from huoguoml.server.service.experiment import ExperimentService


class ExperimentRouter(APIRouter):

    def __init__(self, service: ExperimentService, **extra: Any):
        super().__init__(**extra,
                         prefix="/api/experiments",
                         tags=["experiments"])

        @self.get("", response_model=List[Experiment])
        async def get_experiments():
            return service.get_experiments()

        @self.get("/{experiment_name}", response_model=Experiment)
        async def get_experiment(experiment_name: str):
            experiment = service.get_experiment(experiment_name=experiment_name)
            if experiment is None:
                raise HTTPException(status_code=404)
            return experiment

        @self.get("/{experiment_name}/{experiment_run_nr}", response_model=Run)
        async def get_experiment_run(experiment_name: str, experiment_run_nr: int):
            return service.get_experiment_run(experiment_name=experiment_name,
                                              experiment_run_nr=experiment_run_nr)

        @self.put("/{experiment_name}", response_model=Experiment)
        async def update_experiment(experiment_name: str, experiment: Experiment):
            experiment = service.update_experiment(experiment_name=experiment_name, experiment=experiment)
            if experiment is None:
                raise HTTPException(status_code=404)
            return experiment



import time
from typing import List, Dict, Optional

from sqlalchemy import desc

from huoguoml.schema.ml_model import MLModelIn, MLModel
from huoguoml.schema.ml_model import MLModelTag
from huoguoml.server.entity.ml_model import MLModelORM
from huoguoml.server.entity.ml_service import MLServiceORM
from huoguoml.server.entity.run import RunORM
from huoguoml.server.repository import Repository


class MLModelRepository(Repository):

    def __init__(self, database_url: str, connect_args: Dict):
        super(MLModelRepository, self).__init__(database_url=database_url, connect_args=connect_args)

    def get_ml_models(self):
        session = self.Session()
        return session.query(MLModelORM).order_by(MLModelORM.name).all()

    def get_ml_models_by_name(self, ml_model_name) -> List[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name).order_by(desc(MLModelORM.version)).all()

    def create_ml_model(self, ml_model_in: MLModelIn) -> Optional[MLModelORM]:
        session = self.Session()
        run_orm = session.query(RunORM).filter_by(id=ml_model_in.run_id).first()
        if not run_orm or run_orm and run_orm.ml_model:
            return None

        version = session.query(MLModelORM).filter_by(name=ml_model_in.name).count() + 1
        ml_model = MLModelORM(
            tag=MLModelTag.none.value,
            version="v{}".format(version),
            creation_time=time.time(),
            last_modification=time.time(),
            **ml_model_in.dict())
        session.add(ml_model)
        session.commit()
        session.refresh(ml_model)
        return ml_model

    def get_ml_model_by_name_and_tag(self, ml_model_name: str, tag: int) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   tag=tag).first()

    def get_ml_model_by_name_and_latest(self, ml_model_name: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name).order_by(desc(MLModelORM.version)).first()

    def get_ml_model_files_by_name_and_version(self, ml_model_name: str, ml_model_version: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   version=ml_model_version).first()

    def get_ml_model(self, ml_model_name: str, ml_model_version: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   version=ml_model_version).first()

    def update_ml_model(self, ml_model_name: str, ml_model_version: str, ml_model: MLModel) -> Optional[MLModelORM]:
        session = self.Session()

        ml_model_orm = session.query(MLModelORM).filter_by(name=ml_model_name,
                                                           version=ml_model_version).first()
        if ml_model_orm:
            temp = session.query(MLModelORM).filter_by(name=ml_model_name, tag=ml_model.tag).first()
            if temp and (ml_model.tag == MLModelTag.staging.value or ml_model.tag == MLModelTag.production.value):
                temp.tag = MLModelTag.none.value

            for field, field_value in ml_model.dict(exclude={"id", "run_id", "version", "name"}).items():
                setattr(ml_model_orm, field, field_value)
            session.commit()
            session.refresh(ml_model_orm)

        return ml_model_orm

    def get_ml_model_by_tag(self, ml_model_name: str, tag: str) -> Optional[MLModelORM]:
        session = self.Session()
        return session.query(MLModelORM).filter_by(name=ml_model_name,
                                                   tag=MLModelTag[tag.lower()]).first()

    # TODO: REFACTOR BELOW
    def update_ml_services_by_tag_and_rule(self, ml_model_name: str, model_rule: str, tag: int) -> List[MLServiceORM]:
        session = self.Session()
        ml_services = session.query(MLServiceORM).filter_by(model_name=ml_model_name,
                                                            model_rule=model_rule).all()
        if len(ml_services) == 0:
            return []

        if model_rule == "latest":
            ml_model = self.get_ml_model_by_name_and_latest(ml_model_name=ml_model_name)
        else:
            ml_model = self.get_ml_model_by_name_and_tag(ml_model_name=ml_model_name, tag=tag)

        for ml_service in ml_services:
            ml_service.model_name = ml_model.name
            ml_service.model_version = ml_model.version
            ml_service.model = ml_model
            session.commit()
            session.refresh(ml_service)
        return ml_services



"""
The huoguoml.tracking module provides the options for tracking all experiment runs
"""
import os
from typing import Optional

import requests

from huoguoml.constants import HUOGUOML_DEFAULT_FOLDER
from huoguoml.tracking.client import HuoguoMLAPIRun, HuoguoMLRun, HuoguoMLLocalRun
from huoguoml.util.string import coerce_url


def start_experiment_run(experiment_name: str,
                         artifact_dir: Optional[str] = None,
                         server_uri: Optional[str] = None) -> HuoguoMLRun:
    """
    Start a HuoguoML experiments run and returns it. The return value can be used as a context manager within a with
    block; otherwise, you must call end_experiment_run() to terminate the current run.

    The HugouoML run helps you keep track of your experiment metadata. You can store the metadata in the local file
    system if artifact_dir is specified, or you can send it directly to a HuoguoML tracking server by specifying a
    server_uri.

    Args:
        experiment_name: Name of the experiment under which to create the experiment run
        server_uri: Uri to the HuoguoML server e.g 127.0.0.1:8080
        artifact_dir: The location of the .huoguoml artifact directory for the metadata

    Raises:
        ConnectionError: If Server cannot be found
        ValueError: If server_uri and artifact_dir are both given
    """
    if artifact_dir and server_uri:
        raise ValueError("Either server_rui or artifact_dir must be passed, not both")

    elif server_uri:
        server_uri = coerce_url(server_uri)

        server_res = requests.get(server_uri)
        if server_res.status_code >= 400:
            raise ConnectionError(
                "HuoguoML server cannot be found on {}. Start server with 'huoguoml server' and or try another server "
                "uri".format(
                    server_uri))

        return HuoguoMLAPIRun(server_uri=server_uri, experiment_name=experiment_name)

    elif artifact_dir:
        return HuoguoMLLocalRun(experiment_name=experiment_name,
                                artifact_dir=os.path.join(artifact_dir, HUOGUOML_DEFAULT_FOLDER))



import os

from huoguoml.constants import HUOGUOML_DATABASE_FILE, HUOGUOML_DEFAULT_ZIP_FOLDER


class Service(object):

    def __init__(self, artifact_dir: str):
        self.artifact_dir = os.path.realpath(artifact_dir)
        self.zip_dir = os.path.join(self.artifact_dir, HUOGUOML_DEFAULT_ZIP_FOLDER)

        if not os.path.isdir(self.artifact_dir):
            os.makedirs(self.artifact_dir)
            os.makedirs(self.zip_dir)

        self.database_url = os.path.join("sqlite:///{}".format(artifact_dir), HUOGUOML_DATABASE_FILE)
        self.connect_args = {"check_same_thread": False}



from typing import List, Optional

from huoguoml.schema.ml_service import MLService, MLServiceIn
from huoguoml.server.entity.ml_service import MLServiceORM
from huoguoml.server.repository.ml_service import MLServiceRepository
from huoguoml.server.service import Service


class MLServiceService(Service):

    def __init__(self, artifact_dir: str):
        super(MLServiceService, self).__init__(artifact_dir=artifact_dir)

        self.repository = MLServiceRepository(database_url=self.database_url,
                                              connect_args=self.connect_args)

    def create_or_update_ml_service(self, ml_service_in: MLServiceIn) -> MLService:
        return self.repository.create_or_update_ml_service(ml_service_in=ml_service_in)

    def get_ml_services(self) -> List[MLServiceORM]:
        return self.repository.get_ml_services()

    def get_ml_service(self, service_id: int) -> Optional[MLServiceORM]:
        return self.repository.get_ml_service(service_id=service_id)





from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship, validates

from huoguoml.server.entity import Base


class ExperimentORM(Base):
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True, index=True)
    description = Column(String)

    name = Column(String, index=True, unique=True, nullable=False)
    runs = relationship("RunORM", back_populates="experiment", order_by="desc(RunORM.run_nr)")

    @validates('name')
    def convert_upper(self, key, value):
        self.lower = value.lower()
        return self.lower



import os
import shutil

import uvicorn
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles

from huoguoml.server.api.experiment import ExperimentRouter
from huoguoml.server.api.ml_model import MLModelRouter
from huoguoml.server.api.ml_service import MLServiceRouter
from huoguoml.server.api.run import RunRouter
from huoguoml.server.service.experiment import ExperimentService
from huoguoml.server.service.ml_model import MLModelService
from huoguoml.server.service.ml_service import MLServiceService
from huoguoml.server.service.run import RunService


def start_huoguoml_server(artifact_dir: str, host: str, port: int):
    """
    Starts the HuoguoML server

    Args:
        artifact_dir: Location of the artifact directory
        host: The network address to listen on
        port: The port to listen on
    """

    app = FastAPI()

    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(ExperimentRouter(service=ExperimentService(artifact_dir=artifact_dir)))
    app.include_router(RunRouter(service=RunService(artifact_dir=artifact_dir)))
    app.include_router(MLServiceRouter(service=MLServiceService(artifact_dir=artifact_dir)))
    app.include_router(MLModelRouter(service=MLModelService(artifact_dir=artifact_dir)))

    dashboard_files_dir = os.path.join(os.path.dirname(__file__),
                                       "..",
                                       "dashboard",
                                       "build")
    if os.path.isdir(dashboard_files_dir):
        shutil.copyfile(os.path.join(dashboard_files_dir, "index.html.orig"),
                        os.path.join(dashboard_files_dir, "index.html"))

        with open(os.path.join(dashboard_files_dir, "index.html")) as f:
            newIndexFile = f.read().replace('__HUOGUOML_API_URL__', "http://{}:{}".format(host, port))

        with open(os.path.join(dashboard_files_dir, "index.html"), "w") as f:
            f.write(newIndexFile)

        shutil.copyfile(os.path.join(dashboard_files_dir, "index.html"),
                        os.path.join(dashboard_files_dir, "404.html"))
        shutil.copyfile(os.path.join(dashboard_files_dir, "index.html"),
                        os.path.join(dashboard_files_dir, "405.html"))
        app.mount("/",
                  StaticFiles(directory=dashboard_files_dir, html=True),
                  name="static")

    uvicorn.run(app, host=host, port=port)


if __name__ == '__main__':
    start_huoguoml_server("../../huoguoml-dev/server", "127.0.0.1", 8080)



import os
from itertools import groupby
from operator import attrgetter
from typing import List, Optional

from huoguoml.schema.ml_model import MLModelIn, MLModelRegistry, MLModel, MLModelTag
from huoguoml.server.entity.ml_model import MLModelORM
from huoguoml.server.entity.ml_service import MLServiceORM
from huoguoml.server.repository.ml_model import MLModelRepository
from huoguoml.server.service import Service
from huoguoml.util.zip import create_zip_file


class MLModelService(Service):

    def __init__(self, artifact_dir: str):
        super(MLModelService, self).__init__(artifact_dir=artifact_dir)

        self.repository = MLModelRepository(database_url=self.database_url,
                                            connect_args=self.connect_args)

    def get_ml_models_by_name(self, ml_model_name: str) -> List[MLModelORM]:
        return self.repository.get_ml_models_by_name(ml_model_name=ml_model_name)

    def get_ml_models_groupby_name(self) -> List[MLModelRegistry]:
        ml_models_orm = self.repository.get_ml_models()

        registry_list = []
        for k, g in groupby(ml_models_orm, attrgetter('name')):
            ml_models = list(g)
            registry = MLModelRegistry(
                name=k,
                ml_models=ml_models
            )
            registry_list.append(
                registry
            )
        return registry_list

    def create_ml_model(self, ml_model_in: MLModelIn) -> Optional[MLModelORM]:
        return self.repository.create_ml_model(ml_model_in=ml_model_in)

    def get_ml_model_files_by_name_and_version(self, ml_model_name: str, ml_model_version: str) -> Optional[str]:
        ml_model = self.repository.get_ml_model_files_by_name_and_version(ml_model_name=ml_model_name,
                                                                          ml_model_version=ml_model_version)
        if ml_model:
            run = ml_model.run
            run_dir = os.path.join(self.artifact_dir, run.experiment_name, str(run.run_nr))
            zip_file_path = create_zip_file(src_dir=run_dir, dst_dir=self.zip_dir, zip_name=str(run.id))
            return zip_file_path

    def update_ml_model(self, ml_model_name: str, ml_model_version: str, ml_model: MLModel):
        return self.repository.update_ml_model(ml_model_name=ml_model_name,
                                               ml_model_version=ml_model_version,
                                               ml_model=ml_model)

    def get_ml_model(self, ml_model_name: str, ml_model_version: str) -> Optional[MLModelORM]:
        return self.repository.get_ml_model(ml_model_name=ml_model_name,
                                            ml_model_version=ml_model_version)

    def get_ml_model_by_tag(self, ml_model_name: str, tag: str) -> Optional[MLModelORM]:
        return self.repository.get_ml_model_by_tag(ml_model_name=ml_model_name,
                                                   tag=tag)

    def update_ml_services_by_model_name(self, ml_model_name: str, ml_model_rule: str) -> List[MLServiceORM]:
        return self.repository.update_ml_services_by_tag_and_rule(ml_model_name=ml_model_name,
                                                                  tag=MLModelTag[ml_model_rule],
                                                                  model_rule=ml_model_rule)



import os
from typing import List, Optional

from fastapi import UploadFile

from huoguoml.constants import HUOGUOML_DEFAULT_MODEL_FOLDER, \
    HUOGUOML_METADATA_FILE
from huoguoml.schema.run import Run, RunIn
from huoguoml.server.entity.run import RunORM
from huoguoml.server.repository.run import RunRepository
from huoguoml.server.service import Service
from huoguoml.util.yaml import save_yaml


class RunService(Service):

    def __init__(self, artifact_dir: str):
        super(RunService, self).__init__(artifact_dir=artifact_dir)

        self.repository = RunRepository(database_url=self.database_url,
                                        connect_args=self.connect_args)

    def create_run(self, run_in: RunIn) -> RunORM:
        run_orm = self.repository.create_run(run_in=run_in)
        run_dir = os.path.join(self.artifact_dir, run_orm.experiment_name, str(run_orm.run_nr))
        os.makedirs(run_dir, exist_ok=True)

        save_yaml(yaml_path=os.path.join(run_dir, HUOGUOML_METADATA_FILE),
                  data=Run.from_orm(run_orm).dict())
        return run_orm

    def get_run(self, run_id: int) -> Optional[RunORM]:
        return self.repository.get_run(run_id=run_id)

    def update_run(self, run_id: int, run: Run) -> Optional[RunORM]:
        run = self.repository.update_run(run_id=run_id,
                                         run=run)
        run_dir = os.path.join(self.artifact_dir, run.experiment_name, str(run.run_nr))
        save_yaml(yaml_path=os.path.join(run_dir, HUOGUOML_METADATA_FILE),
                  data=Run.from_orm(run).dict())
        return run

    def update_or_create_run_files(self, run_id: int, files: List[UploadFile]) -> bool:
        run_orm = self.repository.get_run(run_id=run_id)
        if run_orm is None:
            return False

        for file in files:
            file_path = os.path.join(self.artifact_dir,
                                     run_orm.experiment_name,
                                     str(run_orm.run_nr),
                                     HUOGUOML_DEFAULT_MODEL_FOLDER,
                                     file.filename)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, "wb+") as file_object:
                file_object.write(file.file.read())
        return True

    def get_runs(self, experiment_name: Optional[str], run_nrs: Optional[List[int]]) -> List[RunORM]:
        if experiment_name is not None and run_nrs is not None:
            return self.repository.get_runs_by_experiment_name_and_run_nrs(experiment_name=experiment_name,
                                                                           run_nrs=run_nrs)
        else:
            return self.repository.get_runs()



import importlib
import os
import shutil
from typing import List, Any

import requests
from fastapi import APIRouter
from pydantic import create_model

from huoguoml.constants import HUOGUOML_METADATA_FILE, HUOGUOML_DEFAULT_FOLDER
from huoguoml.schema.ml_service import MLService, MLServiceIn
from huoguoml.schema.run import Run
from huoguoml.util.string import concat_uri, coerce_url
from huoguoml.util.yaml import read_yaml
from huoguoml.util.zip import download_and_extract_zip_file


def load_run_model(run: Run):
    module_name = 'huoguoml.tracking.{}'.format(run.model_definition.model_api.module)
    module = importlib.import_module(module_name)
    function = getattr(module, run.model_definition.model_api.name)
    return function(**run.model_definition.model_api.arguments)


class HuoguoMLRouter(APIRouter):
    output_model = None
    input_model = None

    def __init__(self, host: str, port: int, model_name: str, model_rule: str, artifact_dir: str, server_uri: str,
                 prefix: str = "/api",
                 **extra: Any):
        super().__init__(**extra, prefix=prefix)
        server_uri = coerce_url(server_uri)

        ml_service_in = MLServiceIn(host=host, port=port, model_name=model_name, model_rule=model_rule)
        response = requests.post(concat_uri(server_uri, "api", "services"), json=ml_service_in.dict())
        self.ml_service = MLService.parse_raw(response.text)

        self.artifact_dir = artifact_dir
        self.server_uri = server_uri
        self._update()

        @self.post("/update", response_model=MLService)
        async def update_model(updated_ml_service: MLService):
            self.ml_service = updated_ml_service
            self._update()

        @self.post("/predict", response_model=HuoguoMLRouter.output_model)
        async def predict(data: HuoguoMLRouter.input_model):
            return self.model.predict(data)

        @self.get("/version", response_model=MLService)
        async def version():
            return self.ml_service

    def _update(self):
        source_dir = os.getcwd()
        model_url = concat_uri(self.server_uri, "api", "models", self.ml_service.model_name,
                               self.ml_service.model_version)
        model_dir = os.path.join(self.artifact_dir, HUOGUOML_DEFAULT_FOLDER, self.ml_service.model_version)

        if os.path.isdir(model_dir):
            shutil.rmtree(model_dir)
        download_and_extract_zip_file(zip_uri=model_url, dst_dir=model_dir)

        run_dict = read_yaml(os.path.join(model_dir, HUOGUOML_METADATA_FILE))
        run = Run(**run_dict)

        input_model_type = {}
        for input_name, _ in run.model_definition.model_graph.inputs.items():
            input_model_type[input_name] = (List, [])
        HuoguoMLRouter.input_model = create_model('Inputs', **input_model_type)

        output_model_type = {}
        for output_name, _ in run.model_definition.model_graph.outputs.items():
            output_model_type[output_name] = (List, [])
        HuoguoMLRouter.output_model = create_model('Outputs', **output_model_type)

        os.chdir(model_dir)
        self.model = load_run_model(run=run)
        os.chdir(source_dir)
